class MdT:

    def __init__(self, Q, Σ, Γ, δ, qᵢ, qₐ, qᵣ):
        self.Q = Q  # lista di stati
        self.Σ = Σ  # lista di simboli di input
        self.Γ = Γ  # lista di simboli di lavoro
        self.δ = δ  # funzione di transizione. Ritorna una lista di triple
        self.qᵢ = qᵢ  # stato iniziale
        self.qₐ = qₐ  # stato accettante
        self.qᵣ = qᵣ  # stato rifiutante


def esattamente_un_simbolo(M, t):
    formula = ""
    for i in range(-t, t+1):
        if i != -t:
            formula += " ∧ "
        formula += "("
        for j in range(0, t+1):
            if j != 0:
                formula += " ∧ "
            formula += "("
            for (k, σ) in enumerate(M.Γ):
                if k != 0:
                    formula += " ∨ "
                formula += "("
                formula += f"c[{σ},{i},{j}]"
                for γ in M.Γ:
                    if σ != γ:
                        formula += f" ∧ ¬c[{γ},{i},{j}]"
                formula += ")"
            formula += ")"
        formula += ")"
    return formula


def esattamente_uno_stato(M, t):
    formula = ""
    for j in range(0, t+1):
        if j != 0:
            formula += " ∧ "
        formula += "("
        for (k, q) in enumerate(M.Q):
            if k != 0:
                formula += " ∨ "
            formula += "("
            formula += f"e[{q},{j}]"
            for r in M.Q:
                if q != r:
                    formula += f" ∧ ¬e[{r},{j}]"
            formula += ")"
        formula += ")"
    return formula


def esattamente_una_posizione(M, t):
    formula = ""
    for j in range(0, t+1):
        if j != 0:
            formula += " ∧ "
        formula += "("
        for i in range(-t, t+1):
            if i != -t:
                formula += " ∨ "
            formula += "("
            formula += f"p[{i},{j}]"
            for k in range(-t, t+1):
                if k != i:
                    formula += f" ∧ ¬p[{k},{j}]"
            formula += ")"
        formula += ")"
    return formula


def configurazione_iniziale(M, w, t):
    formula = f"e[{M.qᵢ},0] ∧ p[0,0]"
    for i, v in enumerate(w):
        formula += f" ∧ c[{v},{i},0]"
    for i in range(-t, 0):
        formula += f" ∧ c[#,{i},0]"
    for i in range(len(w), t+1):
        formula += f" ∧ c[#,{i},0]"
    return formula


def configurazione_accettante(M, t):
    return f"e[{M.qₐ},{t}]"


def transizione_valida(M, t):
    formula = ""
    for j in range(1, t+1):
        if j != 1:
            formula += " ∧ "
        formula += "("
        for i in range(-t, t+1):
            if i != -t:
                formula += " ∧ "
            formula += "("
            for k, q in enumerate(M.Q):
                if k != 0:
                    formula += " ∧ "
                formula += "("
                for ℓ, σ in enumerate(M.Γ):
                    if ℓ != 0:
                        formula += " ∧ "
                    formula += "("
                    formula += f"p[{i},{j-1}] ∧ e[{q},{j-1}] ∧ c[{σ},{i},{j-1}] ⇒ "
                    for m, tr in enumerate(M.δ(q, σ)):
                        if m != 0:
                            formula += " ∨ "
                        formula += f"(c[{tr[1]},{i},{j}] ∧ e[{tr[0]},{j}] ∧ p[{i+tr[2]},{j}])"
                    formula += ")"
                formula += ")"
            formula += ")"
        formula += ")"
    return formula


def nastro_non_cambia(M, t):
    formula = ""
    for j in range(1, t+1):
        if j != 1:
            formula += " ∧ "
        formula += "("
        for i in range(-t, t+1):
            if i != -t:
                formula += " ∧ "
            formula += f"¬p[{i},{j-1}] ⇒ ("
            for k, σ in enumerate(M.Γ):
                if k != 0:
                    formula += " ∧ "
                formula += f"(c[{σ},{i},{j-1}] ⇔ c[{σ},{i},{j}])"
            formula += ")"
        formula += ")"
    return formula


def MdT_to_formula(M, w, t):
    formula = esattamente_un_simbolo(M, t)
    formula += " ∧ "
    formula += esattamente_una_posizione(M, t)
    formula += " ∧ "
    formula += esattamente_uno_stato(M, t)
    formula += " ∧ "
    formula += configurazione_iniziale(M, w, t)
    formula += " ∧ "
    formula += configurazione_accettante(M, t)
    formula += " ∧ "
    formula += nastro_non_cambia(M, t)
    formula += " ∧ "
    formula += transizione_valida(M, t)
    return formula


def δ(q, σ):
    if q == "q₀" and σ == "a":
        return [("q₀", "a", 1), ("q₂", "b", -1)]
    if q == "q₂":
        return [("q₂", σ, 0)]
    return [("q₁", "#", 1)]


M = MdT(["q₀", "q₁", "q₂"],
        ["a", "b"],
        ["a", "b", "#"],
        δ,
        "q₀",
        "q₂",
        "q₁")
